/* core.c for mdb
 * read info from 'core' file
 * needs MINIX 1.5.6 or later
 * Matthias Pfaller - leo @ marco.de
 * With ideas from  - Philip Murton - philipm@gpu.utcs.utoronto.ca
 *				    - 'adb' by D. Dugger
 *
 */

#include "mdb.h"
#include <fcntl.h>
#include <unistd.h>

PRIVATE long text_size = 0;				/* text size from header*/
PRIVATE long data_size = 0;				/* data size from header*/
PRIVATE int is_sep;						/* Separate I & D 	*/
PRIVATE struct mem_map coremap[NR_SEGS];/* memmap for core	*/
PRIVATE struct proc coreproc;			/* proc-struct for core	*/
PRIVATE int corefd = -1;
PRIVATE off_t seek_text = 0;
PRIVATE off_t seek_data = 0;
PRIVATE off_t seek_stack = 0;

#define CTEXT \
		(sizeof coremap + sizeof coreproc / sizeof (long) * sizeof (long))
#define CDATA	(CTEXT + msize_to_physb(coremap[T].mem_len))
#define CSTACK	(CDATA + msize_to_physb(coremap[D].mem_len))
#define BADADDR(x,seg) \
		((x) < msize_to_physb(coremap[seg].mem_vir) || \
		 (x) > msize_to_physb(coremap[seg].mem_vir + coremap[seg].mem_len))

FORWARD int prerr P_((int req));

PRIVATE int prerr(req)
int req;
{
	switch(req) {
		case T_GETINS:
		case T_SETINS:
			printf("Text address not found");
			return(1);

		case T_GETDATA:
		case T_SETDATA:
			printf("Data address not found");
			return(1);

		default:
			return(0);
	}
}

PUBLIC long myptrace(req, pid, arg1, arg2)
int req;
pid_t pid;
long arg1, arg2;
{
	long ret; int retry;

	if (pid > 1 || req == T_OK) {
		errno = 0;
#if (CHIP == M68000)
		for (retry = 2; retry; retry--) {
			ret = ptrace(req, pid, arg1, arg2);
			if (ret < 0 && errno != 0) {
				switch (req) {
				case T_GETINS:
					req = T_GETDATA;
					continue;
				case T_GETDATA:
					req = T_GETINS;
					continue;
				case T_SETINS:
					req = T_SETDATA;
					continue;
				case T_SETDATA:
					req = T_SETINS;
					continue;
				default:
					break;
				}
			} else
				break;
		}
#else
		ret = ptrace(req, pid, arg1, arg2);
#endif
		if (ret < 0 && errno != 0) {
			errno = EIO;
			if (prerr(req))
				longjmp(mainlp, 1);
			else
				return(-1);
		} else
			return(ret);
	}

	switch (req) {
		default:
			errno = EIO;
			return(-1);

		case T_SETINS:
		case T_SETDATA:
		case T_GETINS:
		case T_GETDATA:
			if ((req == T_GETINS || req == T_SETINS || !is_sep) &&
				 !BADADDR(arg1, T)) {
				arg1 -= msize_to_physb(coremap[T].mem_vir);
				lseek(corefd, seek_text + arg1, SEEK_SET);
			} else if ((req == T_GETDATA || req == T_SETDATA || !is_sep) &&
						!BADADDR(arg1, D)) {
				arg1 -= msize_to_physb(coremap[D].mem_vir);
				lseek(corefd, seek_data + arg1, SEEK_SET);
			} else if ((req == T_GETDATA || req == T_SETDATA || !is_sep) &&
						!BADADDR(arg1, S)) {
				arg1 -= msize_to_physb(coremap[S].mem_vir);
				lseek(corefd, seek_stack + arg1, SEEK_SET);
			} else {
				errno = EIO;
				prerr(req);
				longjmp(mainlp, 1);
			}
			if (req == T_GETINS || req == T_GETDATA)
				read(corefd, (char *) &ret, (unsigned)sizeof(long));
			else
				write(corefd, (char *) &arg2, (unsigned)sizeof(long));
			break;

		case T_GETUSER:
			if ((arg1 & (sizeof(long) - 1)) != 0 ||
			    arg1 > sizeof(struct proc) - sizeof(long)) {
				errno = EIO;
				return(-1);
			}
			ret = *(long *) ((char *) &coreproc + arg1);
			break;

		case T_EXIT:
			close(corefd);
			corefd = -1;
			ret = 0;
			break;
	}
	return(ret);
}

PUBLIC int core_init(corefn, progfn, rw, flat)
char *corefn, *progfn; int rw, flat;
{
	static struct EXEC_T exec;

	if (corefn) {
		if (access(corefn, R_OK)) {
			corefn = NULL;
			printf("No core file\n");
		} else
			if (rw && access(corefn, W_OK)) {
				printf("No write access to %s\n", corefn);
				rw = 0;
			}
	}

	if (flat || !corefn) {
		if (!corefn && rw && !access(progfn, R_OK|W_OK))
			corefd = open(progfn, O_RDWR);
		else
			corefd = open(progfn, O_RDONLY);
		if (corefd == -1) {
			if (rw) {
				printf("Cannot open execfile\n");
				return(core_init(NULL, progfn, 0, 0));
			}
		}
		if (read(corefd, &exec, sizeof(exec)) != sizeof(exec)) {
			close(corefd);
			corefd = -1;
			printf("Bad execfile\n");
			errno = EIO;
			return(-1);
		}
#ifdef MINIX_ST
		coremap[T].mem_vir = 0;
		coremap[T].mem_len = physb_to_msize(exec.a_text + CLICK_SIZE - 1);
		coremap[D].mem_vir = coremap[T].mem_len;
		coremap[D].mem_len = physb_to_msize(exec.a_text +
											exec.a_data + exec.a_bss +
											CLICK_SIZE - 1) -
							 coremap[T].mem_len;
		coremap[S].mem_vir = coremap[T].mem_len + coremap[D].mem_len;
		coremap[S].mem_len = 0;
		if (flat && corefn && !access(corefn, rw?W_OK|R_OK:R_OK)) {
			close(corefd);
			corefd = open(corefn, rw?O_RDWR:O_RDONLY);
		}
		seek_text  = exec.a_hdrlen;
		seek_data  = seek_text + msize_to_physb(coremap[T].mem_len);
		seek_stack = seek_data + msize_to_physb(coremap[D].mem_len);
#endif
#ifdef MINIX_PC
		coremap[T].mem_vir = 0;
		coremap[D].mem_vir = 0;
		if (!is_sep) {
			coremap[T].mem_len = 0;
			coremap[D].mem_len = physb_to_msize(exec.a_text +
											    exec.a_data +
											    CLICK_SIZE - 1);
		}
		coremap[S].mem_vir = coremap[D].mem_len;
		coremap[S].mem_len = 0;
		seek_text  = exec.a_hdrlen;
		seek_data  = seek_text + exec.a_text;
		seek_stack = seek_data + exec.a_data;
#endif
		memset(&coreproc, 0, sizeof(coreproc));
#ifdef MINIX_ST
      /* This doesn't work with Bruce's 386 kernel because sizes are in bytes
       * and the structures have a different number of members.
       * Fortunately coreproc.p_map doesn't seem to be used.
       */
      /* except for setting up saddr... */
		coreproc.p_map[T] = coremap[T];
		coreproc.p_map[D] = coremap[D];
		coreproc.p_map[S] = coremap[S];
#endif
		if (flat && corefn) {
			close(corefd);
			if ((corefd = open(corefn, rw?O_RDWR:O_RDONLY)) == -1) {
				printf("Cannot open corefile\n");
				return(core_init(NULL, progfn, 0, 0));
			}

			seek_text  -= exec.a_hdrlen;
			seek_data  -= exec.a_hdrlen;
			seek_stack -= exec.a_hdrlen;
			return(1);
		}
		else
			return(0);
	}

	if ((corefd = open(corefn, rw?O_RDWR:O_RDONLY)) == -1) {
		printf("Cannot open corefile\n");
		return(core_init(NULL, progfn, 0, 0));
	}

	if (read(corefd, (char *)coremap, (unsigned)sizeof(coremap)) !=
		sizeof(coremap)) {
		close(corefd);
		corefd = -1;
		printf("Bad core file\n");
		return(core_init(NULL, progfn, 0, 0));
	}

#ifdef MINIX_ST			/* text_size already rounded in file (?) */
#if (SHADOWING == 0)
	if (coremap[T].mem_len != 0) 
#else
	if (coremap[T].mem_len != physb_to_msize(text_size))
#endif
#endif
#ifdef MINIX_PC			/* text_size already only partly rounded in file */
						/* It's only rounded to a multple of 16 for separate
						 * I&D.  This is in fact unnecessary, except aligment
						 * to a multiple of 2 or 4 makes the copying for i/o
						 * loading faster.
						 * It's not rounded at all for separate I&D.  But it
						 * should have been rounded to a click boundary for
						 * this case.  As it is, the last click may be part
						 * text and part data.
						 * MINIX_PC make life unnecessarily difficult here
						 * (and elsewhere :-() for common I&D by combining
						 * the text with data and clearing mem_len for text.
						 */
	if (is_sep
		&& coremap[T].mem_len != physb_to_msize(text_size + CLICK_SIZE - 1)
		|| !is_sep && coremap[T].mem_len != 0 /* no fixed lengths in map :( */)
#endif
	{
		close(corefd);
		corefd = -1;
		printf("Bad core magic number\n");
		return(core_init(NULL, progfn, 0, 0));
	}

	if (read(corefd, (char *)&coreproc, (unsigned)sizeof(coreproc)) !=
	    sizeof(coreproc)) {
		close(corefd);
		corefd = -1;
		printf("Bad core file\n");
		errno = EIO;
		return(core_init(NULL, progfn, 0, 0));
	}

	seek_text = CTEXT;
	seek_data = CDATA;
	seek_stack = CSTACK;

	return(corefn?1:0);
}

PUBLIC void prtmap()
{
	printf("%s I & D space\t", (is_sep) ? "Separate " : "Combined ");
	if (curpid == 1)
		printf("File: %s\n", core);
	else if (curpid == 0)
		printf("File: %s\n", prog);
	else
		printf("Pid:	%d\n", curpid);

	printf(" Text: %8lx %8lx\n", saddr, ksize_to_physb(prc->p_map[T].mem_len));
	printf(" Data: %8lx %8lx\n", daddr, ksize_to_physb(prc->p_map[D].mem_len));
	printf("Stack: %8lx %8lx\n", taddr, ksize_to_physb(prc->p_map[S].mem_len));
}

PUBLIC int checkobj(header)
struct EXEC_T *header;
{
	/* Check MAGIC */
	if (header->a_magic[0] != A_MAGIC0 && header->a_magic[1] != A_MAGIC1) {
		printf("Invalid Magic number in header\n");
		return FALSE;
	}

	text_size = header->a_text;
	data_size = header->a_data;
	/* Check type */
	if (header->a_flags == A_EXEC) {
		is_sep = FALSE;
		return TRUE;
	}
	if (header->a_flags == A_SEP) {
		is_sep = TRUE;
		return TRUE;
	}
	printf("Object file not exec\n");
	return TRUE;
}

/*
 *	setMEstart
 *	MOD CMODE TAB4
 *	SYN c
 *	HLP c
 *	TAG
 *	IDC a-zA-Z_0-9\-
 *	PTH /usr/include/sys:/usr/include:
 *	setMEend
 */
