/* This file contains the C startup code for Minix on Intel processors.
 * It cooperates with mpx.x to set up a good environment for main().
 *
 * This code runs in real mode for a 16 bit kernel and may have to switch
 * to protected mode for a 286.
 *
 * For a 32 bit kernel this already runs in protected mode, but the selectors
 * are still those given by the BIOS with interrupts disabled, so the
 * descriptors need to be reloaded and interrupt descriptors made.
 */

#include "kernel.h"
#include "string.h"
#include <minix/boot.h>

PRIVATE char k_environ[256];	/* environment strings passed by loader */

FORWARD _PROTOTYPE( int k_atoi, (char *s) );
FORWARD _PROTOTYPE( char *k_getenv, (char *name) );

/*==========================================================================*
 *				cstart					    *
 *==========================================================================*/
PUBLIC void cstart(parmaddr, parmsize)
char *parmaddr;			/* boot parameters offset */
size_t parmsize;		/* boot parameters length */
{
/* Perform initializations which must be done in real mode. */

  register char *bootp;
  register char *envp;
#if 0
  unsigned machine_magic;
  unsigned mach_submagic;
#endif

  /* Copy the boot parameters to kernel memory. */
  if (parmsize > sizeof k_environ - 2) parmsize = sizeof k_environ - 2;
  for (bootp = &k_environ[0]; parmsize != 0; parmsize--)
	*bootp++ = *parmaddr++;

  /* Just convert environment to boot parameters for now. */
  envp = k_getenv("rootdev");
  if (envp != NIL_PTR) boot_parameters.bp_rootdev = k_atoi(envp);
  envp = k_getenv("ramimagedev");
  if (envp != NIL_PTR) boot_parameters.bp_ramimagedev = k_atoi(envp);
  envp = k_getenv("ramsize");
  if (envp != NIL_PTR) boot_parameters.bp_ramsize = k_atoi(envp);
  envp = k_getenv("scancode");
  if (envp != NIL_PTR) boot_parameters.bp_scancode = k_atoi(envp);
  envp = k_getenv("processor");
  if (envp != NIL_PTR) boot_parameters.bp_processor = k_atoi(envp);

#if 0
  scan_code = boot_parameters.bp_scancode;

  /* Type of VDU: */
  envp = k_getenv("video");
  if (envp != NIL_PTR) {
	if (strcmp(envp, "ega") == 0) ega = TRUE;
	if (strcmp(envp, "vga") == 0) vga = ega = TRUE;
  }

  /* Memory sizes: */
  envp = k_getenv("memsize");
  low_memsize = envp != NIL_PTR ? k_atoi(envp) : 256;
  envp = k_getenv("emssize");
  ext_memsize = envp != NIL_PTR ? k_atoi(envp) : 0;

  /* Determine machine type. */
  processor = boot_parameters.bp_processor;	/* 86, 186, 286, 386, ... */
  machine_magic = get_word(MACHINE_TYPE_SEG, (u16_t *) MACHINE_TYPE_OFF);
  mach_submagic = (machine_magic >> 8) & BYTE;
  machine_magic &= BYTE;
  if (machine_magic == PC_AT) {
	pc_at = TRUE;
	/* could be a PS/2 Model 50 or 60 -- check submodel byte */
	if (mach_submagic == PS_50 || mach_submagic == PS_60)   ps_mca = TRUE;
	if (mach_submagic == PS_50A || mach_submagic == PS_50Z) ps_mca = TRUE;
  } else if (machine_magic == PS_386)
	pc_at = ps_mca = TRUE;
  else if (machine_magic == PS)
	ps = TRUE;

  /* Decide if mode is protected. */
#if INTEL_32BITS
  protected_mode = TRUE;
#else
  protected_mode = processor >= 286 && !using_bios;
#endif
  boot_parameters.bp_processor = protected_mode;	/* FS needs to know */
#endif

  /* Call main() and never return */
  main();
}



/*==========================================================================*
 *				k_atoi					    *
 *==========================================================================*/
PRIVATE int k_atoi(s)
register char *s;
{
/* Convert string to integer - kernel version to avoid bloat from isspace().
 */

  register int total = 0;
  register unsigned digit;
  register int minus = 0;

  while (*s == ' ' || *s == '\t') s++;
  if (*s == '-') {
	s++;
	minus = 1;
  }
  while ((digit = *s++ - '0') < 10) {
	total *= 10;
	total += digit;
  }
  return(minus ? -total : total);
}


/*==========================================================================*
 *				k_getenv				    *
 *==========================================================================*/
PRIVATE char *k_getenv(name)
char *name;
{
/* Get environment value - kernel version of getenv to avoid setting up the
 * usual environment array.
 */

  register char *namep;
  register char *envp;

  for (envp = k_environ; *envp != 0;) {
	for (namep = name; *namep != 0 && *namep == *envp; namep++, envp++)
		;
	if (*namep == '\0' && *envp == '=') return(envp + 1);
	while (*envp++ != 0)
		;
  }
  return(NIL_PTR);
}
