/*
 * This file contains the shared parts for the atari ACSI and SCSI drivers.
 *
 * Each physical drive contains normally up to 4 partitions, but when extended
 * partitioning is used this can be up to 12.
 * each physical drive contains 16 minor devices, so that
 *	DEVICE = (contr << (MINOR_SHIFT+1) | (drive << MINOR_SHIFT | minor;
 * These 16 minor devices per drive are interpreted as follows:
 *	d+0:	whole disk
 *	d+1:	partition 1	according to drive descriptor
 *	d+2:	partition 2	according to drive descriptor
 *	d+3:	partition 3	according to drive descriptor
 *	d+4:	partition 4	according to drive descriptor
 *	d+5:	partition 5	according to drive descriptor
 *	d+6:	partition 6	according to drive descriptor
 *	d+7:	partition 7	according to drive descriptor
 *	d+8:	partition 8	according to drive descriptor
 *	d+9:	partition 9	according to drive descriptor
 *	d+10:	partition 10	according to drive descriptor
 *	d+11:	partition 11	according to drive descriptor
 *	d+12:	partition 12	according to drive descriptor
 *	d+13:	drive descriptor (sector 0)
 *	d+14:	not used 
 *	d+15:	not used
 *
 * The suggested naming scheme is hd followed by the drive number and a 
 * 	letter (a-p) specifiying the partition. 
 *	E.g. /dev/hd1c will correspond with partition 2 on drive 1
 *		drive 1 is the drive at controller 0, lun 1.
 *
 * TODO:
 * - support removable media (hooks are already present)
 *   perhaps: Use PREVENT/ALLOW MEDIUM REMOVAL command after open()/close().
 * - find out which partitions are minix and which not
 * - only access partitions with pi_id == "MIX" (prog to modify pi_id?)
 *   Thereby disallowing a `tosdir/read/write'!
 * - check if retrying is useful. Look at sparc code
 * - wini_hvrdwt:
 *	see if it is possible to issue a request to transfer two
 *	blocks in one command, while programming the dma for only 
 *	one block, and reprogram the dma controller for the next
 *	request afterwards. This seems possible for SCSI, but ACSI?
 *	look also at the sparc code how things are done there.
 * - perhaps: mode sense if command fails
 * - tape code
 */

#include "kernel.h"
#if (MACHINE == ATARI)
#include <minix/com.h>
#include <string.h>
#include "sthdc.h"
#include "proc.h"
#include <sgtty.h>
#include <sys/hdio.h>

#define ADDPARTS 8
#define TRACE(x)	/* x */

/* Parameters for the disk drive. */
#define SECTOR_SIZE	512	/* physical sector size in bytes */
#define NR_DRIVES	(NR_ACSI_DRIVES + NR_SCSI_DRIVES)
#define MINORS_DRIVE	16
#define MINOR_SHIFT	4	/* 2 ** MINOR_SHIFT == #minors/drive */
#define	MAX_MINOR	(NR_DRIVES<<MINOR_SHIFT)
#define MAX_ERRORS	10	/* how often to try rd/wt before quitting */
#define MAX_SEC0_ERRORS	 3	/* how often to try rd/wt sector 0 before */
				/* quitting, must not be too high because */
				/* non existing drives give error messages*/
#define SCAT_SECTORS  (SCSI_SCAT_SECTORS + ACSI_SCAT_SECTORS)
#define CMDSIZ		50	/* maximum command length for ioctl       */
				
PRIVATE int open_count[NR_DRIVES];
PRIVATE struct pi pi[MAX_MINOR];/* begin and size of any minor device */

FORWARD _PROTOTYPE( int do_xfer, (int drive, phys_bytes address,
			long sector, int count, int rw)			);

/*===========================================================================*
 *				wini_open				     *
 *===========================================================================*/
PUBLIC int wini_open(mp)
register message *mp;
{
  /* this function initialises the partition table for a device */
  static struct hi hi;		/* buffer for drive descriptor */
  register int r, maxminor, minor, dev;
  int errors;
 
  dev = mp->DEVICE >> MINOR_SHIFT;
  if (open_count[dev]++ != 0) return(OK);
  /* this function is called too often (on every partition mount) */
  minor = dev << MINOR_SHIFT;
  maxminor = minor + NPARTS + 8;
  /* read sector 0 of the drive */
  /* retry a number of times */
  for (errors = 0; errors < MAX_SEC0_ERRORS; errors++) {
      r = do_xfer(dev, (phys_bytes)&hi, 0L, 1, DEV_READ);
      if (r == OK)
	break;		/* if successful, exit loop */
  }
  /* if the read failed return */
  /* if the read succeeded fill partition table */
  /* note that since pi is in the data segment all entries are */
  /* initialised to 0 (which is also PI_INVALID) */
    
  if (r != OK) {
	open_count[dev]--;
	return(r);
  }
  /* the first minor number is for the whole device */
  pi[minor].pi_start = 0;
  pi[minor].pi_size = hi.hd_size;
  pi[minor].pi_flag = PI_VALID;
  
  for (r = 0, minor++; (r < NPARTS) && (minor <= maxminor); r++) {
      if (strncmp(hi.hd_pi[r].pi_id, "XGM", 3) == 0)
      {	/* yikes, an extended partition */
      	/* only one extended partition can occur in the root sector */
	struct hi hiext; /* buffer for extended part. drive descriptor */
	register int s, ext; /* ext gives the number of the extended part. */
	long offset;	 /* where the next extended root sector is */
	
        ext = r;
	hiext.hd_pi[ext].pi_start = 0;
	/* extended partitions contain an extended root sector
	   which is filled with one regular, and zero or one
	   extended partitions. 
	   The start of all extended partitions is relative to the
	   extended partition in the real root sector
          */
          while (ext >= 0)
          {
	    offset = hi.hd_pi[r].pi_start + hiext.hd_pi[ext].pi_start;
	    ext = -1;
	    /* retry a number of times */
	    for (errors = 0; errors < MAX_SEC0_ERRORS; errors++) {
	        s = do_xfer(dev, (phys_bytes)&hiext,
	                    offset, 1, DEV_READ);
	        if (s == OK)
		    break;		/* if successful, exit loop */
	    }
	    for (s = 0; (s < NPARTS) && (minor <= maxminor); s++) {
	    	if (hiext.hd_pi[s].pi_flag == PI_VALID)
	    	{
	            if (strncmp(hiext.hd_pi[s].pi_id, "XGM", 3) == 0)
	                ext = s;
		    else 
		    {
			if (strncmp(hiext.hd_pi[s].pi_id, "GEM", 3) != 0 &&
			    strncmp(hiext.hd_pi[s].pi_id, "BGM", 3) != 0 &&
			    strncmp(hiext.hd_pi[s].pi_id, "MIX", 3) != 0 )
					pi[minor].pi_flag = PI_INVALID;
			else {
			    pi[minor] = hiext.hd_pi[s];
			    pi[minor].pi_start += offset;
			    /* some sanity checks */
			    if ((pi[minor].pi_start + pi[minor].pi_size - 1 >
			           hi.hd_pi[r].pi_start + hi.hd_pi[r].pi_size) ||
			        (pi[minor].pi_start < hi.hd_pi[r].pi_start) ||
			        (pi[minor].pi_size < 0))
				    pi[minor].pi_flag = PI_INVALID;
			}
			minor++;
		    }
	    	}
	    }
	}
      } else {
	if (strncmp(hi.hd_pi[r].pi_id, "GEM", 3) != 0 &&
	    strncmp(hi.hd_pi[r].pi_id, "BGM", 3) != 0 &&
	    strncmp(hi.hd_pi[r].pi_id, "MIX", 3) != 0 )
			pi[minor].pi_flag = PI_INVALID;
	else
			pi[minor] = hi.hd_pi[r];
	minor++;
      }
  }
#if (ADDPARTS != 0)
  for (r = 0; (r < ADDPARTS) && (minor <= maxminor); r++, minor++){
	/* It is better to check. The danger is too big, in case you
	 * boot with a wrong diskette. (jw)
	 */
	if (strncmp(hi.hd_addpi[r].pi_id, "GEM", 3) != 0 &&
	    strncmp(hi.hd_addpi[r].pi_id, "BGM", 3) != 0 &&
	    strncmp(hi.hd_addpi[r].pi_id, "MIX", 3) != 0 )
			pi[minor].pi_flag = PI_INVALID;
	else
			pi[minor] = hi.hd_addpi[r];
  }
#endif
  /* the next partition is for sector 0 */
  pi[maxminor + 1].pi_size = 1;
  pi[maxminor + 1].pi_flag = PI_VALID;

  minor = dev << MINOR_SHIFT;
  for (r = 0; r < MINORS_DRIVE; r++, minor++)
     if (pi[minor].pi_flag != PI_INVALID)
     {
	    
	/* some sanity checks */
        if ((pi[minor].pi_start + pi[minor].pi_size -1 > hi.hd_size) ||
	    (pi[minor].pi_start < 0) ||
	    (pi[minor].pi_size < 0))
		pi[minor].pi_flag = PI_INVALID;
  	TRACE(
		printf("hd%d%c: 1st=%D, tot=%D\n", minor >> MINOR_SHIFT,
			(minor % MINORS_DRIVE) + 'a',
			pi[minor].pi_start,
			pi[minor].pi_size
		);
	);
      }
  return(OK);
}

/*===========================================================================*
 *				wini_rdwt				     *
 *===========================================================================*/
PUBLIC int wini_rdwt(mp)
register message *mp;
{
  return(wini_transfer( mp->m_type, mp->PROC_NR, mp->DEVICE, mp->POSITION,
		        mp->COUNT, (vir_bytes) mp->ADDRESS));
}

#if (SCAT_SECTORS > 0)
/*===========================================================================*
 *				wini_hvrdwt				     *
 *===========================================================================*/
PUBLIC int wini_hvrdwt(mp)
register message *mp;
{
/* Fetch a vector of i/o requests and handle them for hard disk. Use an inter-
 * nal buffer for contiguous requests to avoid overhead for single block dma
 * requests from FS and keep up with iterleave 1:1 for non-caching disk con-
 * trollers (makes io upto 3 times faster). Return status in the user vector.
 */

  register struct iorequest_s *iop;
  static struct iorequest_s iovec[NR_BUFS];
  phys_bytes iovec_phys, usrvec_phys, phys_ptr;
  unsigned nr_requests;
  static char buff[SCAT_SECTORS * SECTOR_SIZE];
  char *buf;
  int bufsiz;
  register long startpos, endpos;
  register unsigned request;
  register int result, limit, current;

  if (mp->DEVICE < (NR_ACSI_DRIVES <<MINOR_SHIFT))
  {
	buf = buff;
	bufsiz = ACSI_SCAT_SECTORS * SECTOR_SIZE;
  }
  else
  {
	buf = &buff[ACSI_SCAT_SECTORS * SECTOR_SIZE];
	bufsiz = SCSI_SCAT_SECTORS * SECTOR_SIZE;
  }
  nr_requests = mp->COUNT;
  if (nr_requests > sizeof iovec / sizeof iovec[0])
	panic("FS gave hd driver too big an i/o vector", nr_requests);
  iovec_phys = umap(proc_ptr, D, (vir_bytes) iovec, (vir_bytes) sizeof iovec);
  usrvec_phys = numap(mp->PROC_NR, (vir_bytes) mp->ADDRESS,
		      (vir_bytes) (nr_requests * sizeof iovec[0]));
  if (usrvec_phys == 0)
	panic("FS gave hd driver bad i/o vector", (int) mp->ADDRESS);
  phys_copy(usrvec_phys, iovec_phys, (phys_bytes)nr_requests * sizeof iovec[0]);

  iop = iovec;
  startpos = endpos = iop->io_position;
  for(current = 0, limit = 1; current < nr_requests; limit = current + 1) {
	request = iop->io_request & ~OPTIONAL_IO;
	if (request == DEV_READ || request == DEV_WRITE) {
	    /* Handle all requests fitting in buf at once. Read requests should
	     * not skip more then 4 sectors. Write requests must be contiguous.
	     */
	    startpos = endpos = iop->io_position;
	    for (limit = current; limit < nr_requests; limit++, iop++) {
		if ((iop->io_request & ~OPTIONAL_IO) != request		    ||
		    iop->io_position+iop->io_nbytes > startpos+bufsiz       ||
		    (request == DEV_WRITE && iop->io_position != endpos)    ||
		    (request == DEV_READ &&
		    iop->io_position >= endpos + 4*SECTOR_SIZE))
			break;
		/* Copy to buffer on write requests */
		if (request == DEV_WRITE) {
		    if ((phys_ptr = numap(mp->PROC_NR, (vir_bytes) iop->io_buf,
					  (vir_bytes) iop->io_nbytes)) == 0)
			break;	/* Something wrong with user's address */
		    phys_copy( phys_ptr,
			(phys_bytes) &buf[iop->io_position-startpos],
			(phys_bytes) iop->io_nbytes );
		}
		endpos = iop->io_position + iop->io_nbytes;
	    }
	    if (limit == current) limit++;
	    iop = &iovec[current];
	}
	if (limit > current+1) {
	    result = wini_transfer( request, WINCHESTER, mp->DEVICE, startpos,
				    (int) (endpos-startpos), (vir_bytes) buf);
	    if (result == endpos-startpos) {
		for (; current < limit; iop++, current++) {
		    result = 0;
		    if (request == DEV_READ) {
			if ((phys_ptr=numap(mp->PROC_NR, (vir_bytes)iop->io_buf,
					    (vir_bytes) iop->io_nbytes)) == 0)
				result = EINVAL;	/* user addr wrong */
			else phys_copy(
				(phys_bytes) &buf[iop->io_position - startpos],
				phys_ptr, (phys_bytes) iop->io_nbytes );
		    }
		    iop->io_nbytes = result;
		}
	    } else			/* do single transfers on error. */
		limit = current+1;	/* may cause additional transfers ! */
	}
	if (limit == current+1) {
	    result = wini_transfer( iop->io_request & ~OPTIONAL_IO, mp->PROC_NR,
				    mp->DEVICE, iop->io_position, 
				    (int) iop->io_nbytes,
				    (vir_bytes) iop->io_buf);
	    if (result == 0) break;	/* EOF */
	    if (result < 0) {
		iop->io_nbytes = result;
		if (iop->io_request & OPTIONAL_IO) break;  /* abort if opt */
	    } else
		iop->io_nbytes -= result;
	    current++; limit++; iop++;
	}
  }

  phys_copy(iovec_phys, usrvec_phys, (phys_bytes)nr_requests * sizeof iovec[0]);
  return(OK);
}
#endif

/*===========================================================================*
 *				do_xfer					     *
 *===========================================================================*/
PRIVATE int do_xfer(drive, address, sector, count, rw)
int drive;
phys_bytes address;
long sector;
int count;
int rw;
{
  int	r;
  unsigned char cmd[6];
  int lun;
#ifdef _ANSI
  int (*do_cmd) (int drive,  unsigned char *cmd, int cmdlen,
		phys_bytes address, phys_bytes data_len,  int rw);
#else
  int (*do_cmd) ();
#endif

  TRACE(printf("hd drive:%d address:0x%lx sector:%ld count=%d cmd:%s\n",
	drive, address, sector, count, (rw==DEV_READ)?"READ":"WRITE")
  );
  if (drive < NR_ACSI_DRIVES)
  {
	lun = drive & 1;
        cmd[0] = (drive>>1)<<5;
	do_cmd = acsi_cmd;
  }
  else
  {
	lun = 0;
	cmd[0] = 0;
	drive = drive - NR_ACSI_DRIVES;
	do_cmd = scsi_cmd;
  }
  if (rw == DEV_READ) {
        cmd[0] = cmd[0] | HD_RD;
  } else {
        cmd[0] = cmd[0] | HD_WR;
  }
  cmd[1] = (lun << 5) | (((int)(sector>>16))&0x1F);
  cmd[2] = ((int)sector >> 8) & 0xff;
  cmd[3] = sector & 0xff;
  cmd[4] = count & 0xff;
  cmd[5] = 0;
  r = do_cmd(drive, cmd, sizeof(cmd), address,
		(phys_bytes)count * SECTOR_SIZE, rw);
  if (r != OK)
  {
	char buf[16];
	phys_bytes addr;
	int i;

	cmd[0] = (cmd[0] & 0xe0) | HD_RQS;
  	cmd[1] = (lun << 5);
	cmd[2] = 0;
	cmd[3] = 0;
	cmd[4] = 4;
	cmd[5] = 0;
	addr = umap(proc_ptr, D, (vir_bytes)buf, (vir_bytes) 16);
  	do_cmd(drive, cmd, sizeof(cmd), addr, (phys_bytes)16, rw);
	printf("sense returned:\n");
	for (i=0; i < 16; i++) printf("%x ", buf[i]);
	printf("\n");
  }
  return(r);
}


/*===========================================================================*
 *				wini_transfer				     *
 *===========================================================================*/
PUBLIC int wini_transfer(rw, pnr, minor, pos, count, vadr)
int rw;				/* DEV_READ, DEV_WRITE */
int pnr;			/* Process number */
register int minor;		/* minor device number affected */
long pos;			/* position off data on device */
register int count;		/* byte count to transfer */
vir_bytes vadr;			/* virtual src/dst of user buffer */
{
  register struct proc	*rp;
  register long		secnum, avail;
  register phys_bytes	address;
  int			r, errors, drive;

  drive = minor >> MINOR_SHIFT;
  if (drive < 0 || drive >= NR_DRIVES)
	return(EIO);
  if (pi[minor].pi_flag == PI_INVALID)
	return(EIO);
  if ((pos % SECTOR_SIZE) != 0)
	return(EINVAL);
  secnum = (long)(pos / SECTOR_SIZE);
  if ((count % SECTOR_SIZE) != 0)
	return(EINVAL);
  rp = proc_addr(pnr);
  address = umap(rp, D, vadr, (vir_bytes) count);
  if (address == 0)
	return(EINVAL);
  count /= SECTOR_SIZE;
  TRACE(printf("hd%d: %s: sec=%D; cnt=%d\n",
	minor, rw == DEV_READ ? "read" : "write", secnum, count));
  avail = pi[minor].pi_size - secnum;
  if (avail <= 0)
	return(0);
  if (avail < count)
	count = avail;
  if (count <= 0)
	return(0);
  secnum += pi[minor].pi_start;
  rp->p_physio = 1;		/* disable (un)shadowing */
  /* This loop allows a failed operation to be repeated. */
  for (errors = 0; errors < MAX_ERRORS; errors++) {
	r = do_xfer(drive, address, secnum, count, rw);
	if (r == OK)
		break;		/* if successful, exit loop */
  }
  rp->p_physio = 0;		/* enable (un)shadowing */
  if (r != OK)
	return(EIO);
  return(count * SECTOR_SIZE);
}

/*===========================================================================*
 *				wini_close				     *
 *===========================================================================*/
PUBLIC int wini_close(mp)
register message *mp;
{
  /* this function initialises the partition table for a device */
  register int minor, r;
  register int dev;

  dev = mp->DEVICE >> MINOR_SHIFT;
  if (open_count[dev] == 0) {
  	printf("hd: Closing non open device %d\n", mp->DEVICE);
  	return(OK);
  }
  /* if mp->COUNT != 0 then the caller still has another pointer to the
     device (created by a dup or fork call), so don't do the final close */
  if (mp->COUNT != 0) return(OK);
  if (--open_count[dev] != 0) return(OK);
  minor = dev << MINOR_SHIFT;
  for (r = 0; r < MINORS_DRIVE; r++, minor++)
     pi[minor].pi_flag = PI_INVALID;
  return(OK);
}

PUBLIC int wini_ioctl(mp)
register message *mp;
{
  register phys_bytes	uaddress, kaddress;
  register struct proc	*rp;
  register int r, drive, rw;
  struct   scsi_cmd kcmd;
  /* The acsi controller needs that the data buffer is in ST ram. Therefore
     all ioctls are done through the buf array. Requests greater than
     SECTOR_SIZE are rejected.
  */
  static   _VOLATILE unsigned char buf[SECTOR_SIZE];
  unsigned char cmd_buf[CMDSIZ];

  rp = proc_addr(mp->PROC_NR);
  drive = mp->DEVICE >> MINOR_SHIFT;

  switch(mp->TTY_REQUEST) {
     case SCSI_WRITE:
     case SCSI_READ:
	  uaddress = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) sizeof(struct scsi_cmd));
	  break;
     default:
	  return(EINVAL);
  }
  if (uaddress == 0)
	return(EINVAL);
  rw = DEV_READ;
  switch(mp->TTY_REQUEST) {
     case SCSI_WRITE:
	  rw = DEV_WRITE;
     case SCSI_READ:
	  /* copy the command block */
	  kaddress = umap(proc_ptr, S, (vir_bytes) &kcmd, (vir_bytes) sizeof(kcmd));
	  phys_copy(uaddress, kaddress, (vir_bytes) sizeof(kcmd));

	  /* copy the command bytes */
	  uaddress = umap(rp, D, (vir_bytes) kcmd.cmd, (vir_bytes) kcmd.cmd_cnt);
	  if (uaddress == 0)
		return(EINVAL);
	  if (kcmd.cmd_cnt > CMDSIZ)
		return(EINVAL);
	  kaddress = umap(proc_ptr, S, (vir_bytes) cmd_buf, (vir_bytes) kcmd.cmd_cnt);
	  phys_copy(uaddress, kaddress, (vir_bytes) kcmd.cmd_cnt);
     	  switch(buf[0]) {	/* test command */
	    case HD_TUR:
	    case HD_RQS:
	    case HD_INQ:
	    case HD_MSN:
			/* non-destructive commands are allowed on all minors */
			break;
	    default:
		/* destructive commands are only allowed on minor 0 */
     	  	if ((mp->DEVICE & 15) != 0)
     	  		return(EINVAL);
	  }

	  uaddress = umap(rp, D, (vir_bytes) kcmd.buf, (vir_bytes) kcmd.size);
	  if (uaddress == 0)
		return(EINVAL);

	  if (kcmd.size > SECTOR_SIZE)
		return(EINVAL);


	  if (drive < NR_ACSI_DRIVES)
	  {
		kaddress = umap(proc_ptr, S, (vir_bytes) buf, (vir_bytes) sizeof(buf));
		cmd_buf[0] = (drive>>1)<<5 | cmd_buf[0];
	        cmd_buf[1] = ((drive & 1) << 5) | cmd_buf[1];
	        if (rw == DEV_WRITE)
	        {
	       		phys_copy(uaddress, kaddress, (vir_bytes) kcmd.size);
	        }
	        r = acsi_cmd(drive, cmd_buf, kcmd.cmd_cnt, kaddress, kcmd.size, rw);
	        if (rw != DEV_WRITE)
	        {
		      phys_copy(kaddress, uaddress, (vir_bytes) kcmd.size);
	        }
	  }
	  else
	  {
		drive = drive - NR_ACSI_DRIVES;
	        r = scsi_cmd(drive, cmd_buf, kcmd.cmd_cnt, uaddress, kcmd.size, rw);
	  }
	  break;
  }
  return(r);
}
#endif
